# copyright (C) 1997-1999 Jean-Luc Fontaine (mailto:jfontain@free.fr)
# this program is free software: please read the COPYRIGHT file enclosed in this package or use the Help Copyright menu

set rcsId {$Id: ping.tcl,v 1.4.1.4 1999/10/02 20:59:09 jfontain Exp $}


package provide ping 1.0
package require network 1.4
package require stooop 3.7
namespace import stooop::*


namespace eval ping {

    variable void 000000

    array set data {
        updates 0
        0,label host 0,type dictionary 0,message {host or gateway name} 0,anchor left
        1,label address 1,type dictionary 1,message {IP address} 1,anchor left
        2,label replied 2,type dictionary 2,message {reply address} 2,anchor left
        3,label period 3,type integer 3,message {polling period in seconds}
        4,label count 4,type integer 4,message {number of packets transmitted for each period}
        5,label size 5,type integer 5,message {packet size in bytes}
        6,label timeout 6,type integer 6,message {maximum time spent waiting for a response for each packet, in seconds}
        7,label delay 7,type integer 7,message {delay between transmitted packets in seconds}
        8,label transmitted 8,type integer 8,message {number of packets transmitted for the last period}
        9,label received 9,type integer 9,message {number of packets received for the last period}
        10,label loss 10,type integer 10,message {percentage of packets lost for the last period}
        11,label minimum 11,type integer 11,message {minimum round trip time for the last period, in milliseconds}
        12,label averaged 12,type integer 12,message {averaged round trip time for the last period, in milliseconds}
        13,label maximum 13,type integer 13,message {maximum round trip time for the last period, in milliseconds}
        views {
            {visibleColumns {0 1 3 4 5 6 7} sort {0 increasing}}
            {visibleColumns {0 2 8 9 10 11 12 13} sort {0 increasing}}
        }
        switches {-f 1 -r 1}
    }
    set file [open [file join $::packageDirectory(ping) ping.htm]]
    set data(helpText) [read $file]                                                           ;# initialize HTML help data from file
    close $file

    proc initialize {optionsName} {
        upvar $optionsName options
        variable data
        variable void

        if {[catch {set name $options(-f)}]} {                                                            ;# no hosts file specified
            set name [file join $::packageDirectory(ping) hosts]                                                      ;# use default
        }
        set file [open $name]
        set period 2147483647
        set row 0
        while {[gets $file line]>=0} {
            set line [string trim $line]
            if {[string match #* $line]} continue                                                                   ;# skip comments
            foreach "data($row,0) data($row,3) data($row,6) data($row,4) data($row,5) data($row,7)" $line {}
            ### do some error checking here on host, period>0, size>64, ... ###
            set data($row,4) 1                                                                             ;# forced if this release
            set data($row,7) 0                                                                             ;# unused if this release
            array set data [list\
                $row,1 [::network::addressfromhost $data($row,0)] $row,2 {}\
                $row,8 $void $row,9 $void $row,10 $void $row,11 $void $row,12 $void $row,13 $void\
            ]
            if {$data($row,3)<$period} {
                set period $data($row,3)                                           ;# keep track of the minimum period for all hosts
            }
            incr row
        }
        close $file
        set data(pollTimes) -$period

        set requestsHost 127.0.0.1                                                                          ;# local host by default
        catch {set requestsHost $options(-r)}                                                   ;# may be overridden in command line
        ::session::initialize $requestsHost ::ping::received

        set rows $row
        for {set row 0} {$row<$rows} {incr row} {                                                              ;# initialize polling
            launch $row
        }
    }

    proc launch {row} {
        variable data

        new session $data($row,1) $data($row,4) $data($row,6) $data($row,5) "::ping::process $row"
    }

    proc process {row session replyAddress time} {
        variable data
        variable void

        delete $session
        if {$time<0} {
           array set data [list $row,2 {} $row,8 1 $row,9 0 $row,10 100 $row,11 $void $row,12 $void $row,13 $void]
        } else {
           array set data [list $row,2 $replyAddress $row,8 1 $row,9 1 $row,10 0 $row,11 $time $row,12 $time $row,13 $time]
        }
        after [expr {$data($row,3)*1000}] "::ping::launch $row"                           ;### time pinging should be subtracted ###
    }

    proc received {} {
        variable event

        if {![info exists event]} {      ;# optimize by waiting a short time before updating so that multiple updates can be grouped
            set event [after 1000 ::ping::update]
        }
    }

    proc update {} {
        variable event
        variable data

        unset event
        incr data(updates)
    }

}


class session {

    proc session {this address count timeout size command} {
        variable socket

        set ::session::($this,command) $command
        puts -nonewline $socket [binary format c4Ic4ccccSS {0 1 0 0} $this [split $address .] 255 $timeout 0 0 $size 0]
        flush $socket
    }

    proc ~session {this} {}

    proc initialize {requestsHost receiveCommand} {                             ;# command to be invoked once reception is completed
        variable socket
        variable command

        variable socket [socket $requestsHost icmp]
        fconfigure $socket -blocking 0 -translation binary
        set command $receiveCommand
        fileevent $socket readable ::session::receive
    }

    proc receive {} {
        variable socket
        variable command

        while {[binary scan [read $socket 16] xxcxIc4I status session bytes time]==4} {
            catch {unset address}
            foreach byte $bytes {
                if {[info exists address]} {
                    append address .
                }
                append address [expr {($byte+256)%256}]                                                       ;# convert to unsigned
            }
            uplevel #0 $::session::($session,command) $session $address $time
        }
        uplevel #0 $command
    }

}
